//
//  PublishCommand.swift
//  EscapeCore
//
//  Created by Salma ElTarzi on 8/6/18.
//

import Foundation
import SwiftCLI
import SwiftShell
import FileUtils
import PerfectMarkdown

class RNPublishCommand: Command {
    let name = "publish"
    let shortDescription = "Publish the react native Instabug library to npm, creates a release on Github along with the changelog and updates the dashboard's changelog. The following environment variables are required for the command to run: RELEASE_GITHUB_TOKEN, DASHBOARD_TOKEN, NPM_TOKEN"
    
    func execute() throws {
        do {
            try publishToGithubAndDashboard()
        } catch {
            print(error)
            print("Publishing to Github and/or dashboard failed");
            exit(1)
        }
        do {
            try publishToNpm()
        } catch {
            print(error)
            print("Publishing to npm failed");
            exit(1)
        }
        
    }
    
    
    private func publishToGithubAndDashboard() throws {
        let owner = "Instabug";
        let repository = "Instabug-React-Native";
        var version = "";
        let changeLog = Helpers.getChangelog();
        let gitToken = ProcessInfo.processInfo.environment["RELEASE_GITHUB_TOKEN"]!;
        let dashboardToken = ProcessInfo.processInfo.environment["DASHBOARD_TOKEN"]!;
        
        // Getting Version number from package.json
        let pathVersion = "package.json"
        let dataVersion = try NSString(contentsOfFile: pathVersion,
                                       encoding: String.Encoding.ascii.rawValue)
        
        let matched = Helpers.matches(for: "\"version\": \"([0-9]-*.*)+[0-9]\"", in: dataVersion as String)
        let grepVersion = Helpers.matches(for: "([0-9]{1,}\\.)+[0-9]{1,}", in: matched! as String)
        version = "v" + grepVersion!;
        
        // Releases as bullet points seperated by line breaks (for github release)
        var gitReleaseNotes = "";
        var markdownChangelog = "";
        for change in changeLog {
            gitReleaseNotes += change + "\\n";
            markdownChangelog += change + "\n";
        }
        markdownChangelog += "\n";
        
        // Publish tag and Release to Github
        try "".runAndPrint("curl", "--data", "{\"tag_name\": \"\(version)\", \"target_commitish\": \"master\", \"name\": \"Release \(version)\", \"body\": \"\(gitReleaseNotes)\", \"draft\": false, \"prerelease\": false}", "https://api.github.com/repos/\(owner)/\(repository)/releases?access_token=\(gitToken)");
        
        // Converting markdown to HTML for dashboard changelog
        guard let dashboardChangelogs = markdownChangelog.markdownToHTML else {
            return ;
        }
        let dashChangelog = String(dashboardChangelogs.filter { !"\n".contains($0) });
        var dashboardChangelog = "";
        for char in dashChangelog {
            if (char == "\"") {
                dashboardChangelog += "\\";
            }
            dashboardChangelog += String(char);
        }
    
        // Getting today's date to be published on the dashboard
        let date = Date();
        let formatter = DateFormatter();
        formatter.dateFormat = "MMM dd yyyy";
        let today = formatter.string(from: date);
        
        // Publishing Release and Changelog to dashboard
        try "".runAndPrint("curl", "-X", "POST", "-H", "content-type: application/json", "-f", "-d", "{\"token\": \"\(dashboardToken)\",\"change_log\": {\"version\": \"\(version)\",\"release_date\": \"\(today)\",\"target_os\": \"4\",\"sdk_changes\": \"\(dashboardChangelog)\"},\"sdk_remote_url\": \"https://github.com/Instabug/instabug-reactnative\",\"sample_app_remote_url\": \"https://github.com/Instabug/instabug-reactnative/tree/master/InstabugSample\"}", "https://api.instabug.com/api/internal/change_logs");
        
    }
    
    
    private func publishToNpm() throws {
        // Login to npm
        let token = ProcessInfo.processInfo.environment["NPM_TOKEN"]!;
        let command = main.run(bash: "npm set //registry.npmjs.org/:_authToken=\(token)");
        if command.error != nil {
            Logger.printError(command.stderror);
        } else {
            Logger.printSuccess("Logged in to npm with username instabug");
        }
        
        // Publish to npm
        try "".runAndPrint("npm", "publish", "./");
    }
}

class PublishCommand: Command {
    let name = "publish"
    let shortDescription = "Publish the xamarin nuget packages for instabug and updates the dashboard's changelog. The following environment variables are required for the command to run: DASHBOARD_TOKEN"

    init() {
        
    }
    
    let projectDirectory = main.currentdirectory;
    
    func execute() throws {
        setApiKey()

        // Installing dependencies of Android Modules
        let androidModules = Constants.packageIds.androidPackages
        for(packageId, androidModule) in androidModules {
            installNugetPackages(for: androidModule, project: projectDirectory)
        }

        // Installing dependencies, packing and publishing of Android Main Module
        pack(for: Constants.xamariniAndroidMainModule, project: projectDirectory)
        publish(module:Constants.xamariniAndroidMainModule, packageID:Constants.packageIds.mainAndroid)

        // Installing dependencies, packing and publishing of iOS Module
        pack(for: Constants.xamariniOSModule, project: projectDirectory)
        publish(module:Constants.xamariniOSModule, packageID:Constants.packageIds.iOS)

        do {
            try publishToDashboard()
        } catch {
            print(error)
            print("Publishing to dashboard failed");
            exit(1)
        }
    }
    
    private func publishToDashboard() throws {
        main.currentdirectory = projectDirectory;
        var version = "";
        let changeLog = Helpers.getChangelog();
        let dashboardToken = ProcessInfo.processInfo.environment["DASHBOARD_TOKEN"]!;
        
        let module = Constants.xamariniOSModule;
        let moduleDirectory = projectDirectory.appending("\(module)/\(module)")
        //main.currentdirectory = moduleDirectory
        let nuspecFileContents = Helpers.nuspecFileContents(from: moduleDirectory.appending("/\(module).nuspec"))
        
        let versionNumberTag = Helpers.versionNumber(from: nuspecFileContents)
        version = "v" + Helpers.versionNumberString(from: versionNumberTag)
        
        print(version)
        
        // Releases as bullet points seperated by line breaks (for github release)
        var gitReleaseNotes = "";
        var markdownChangelog = "";
        for change in changeLog {
            gitReleaseNotes += change + "\\n";
            markdownChangelog += change + "\n";
        }
        markdownChangelog += "\n";
        
        // Converting markdown to HTML for dashboard changelog
        guard let dashboardChangelogs = markdownChangelog.markdownToHTML else {
            return ;
        }
        let dashboardChangelog = String(dashboardChangelogs.filter { !"\n".contains($0) });
        
        
        // Getting today's date to be published on the dashboard
        let date = Date();
        let formatter = DateFormatter();
        formatter.dateFormat = "MMM dd yyyy";
        let today = formatter.string(from: date);
        
        // Publishing Release and Changelog to dashboard
        try "".runAndPrint("curl", "-X", "POST", "-H", "content-type: application/json", "-f", "-d", "{\"token\": \"\(dashboardToken)\",\"change_log\": {\"version\": \"\(version)\",\"release_date\": \"\(today)\",\"target_os\": \"5\",\"sdk_changes\": \"\(dashboardChangelog)\"},\"sdk_remote_url\": \"\",\"sample_app_remote_url\": \"https://github.com/Instabug/instabug-reactnative/tree/master/InstabugSample\"}", "https://api.instabug.com/api/internal/change_logs");
        
    }
    
    
    /**
     Publishes a package for a certain module.
     
     - Parameters:
     - module: The module being published.
     - packageID: The id of the package being published.
     */
    func publish(module: String, packageID: String) {
        
        // Navigate to the directory where the .nupkg file is located
        let moduleDirectory = projectDirectory.appending("\(module)/\(module)")
        main.currentdirectory = moduleDirectory
        do {
            let nuspecFileContents = Helpers.nuspecFileContents(from: moduleDirectory.appending("/\(module).nuspec"))
            
            let versionNumberTag = Helpers.versionNumber(from: nuspecFileContents)
            let versionNumber = Helpers.versionNumberString(from: versionNumberTag)
            try "".runAndPrint("nuget", "push", "\(packageID).\(versionNumber).nupkg", "-Source", "https://api.nuget.org/v3/index.json");
            Logger.printSuccess("Successfully published \(module) package!")
        } catch {
            Logger.printFatalError("Failed to publish \(module) package to Nuget!")
            exit(1)
        }
    }

    private func installNugetPackages(for module: String, project projectDir: String) {
        main.currentdirectory = projectDir
        main.run(bash: "nuget install " + module + "/" + module + "/packages.config -OutputDirectory " + module + "/Packages")
    }
    
    private func pack(for module: String, project projectDir: String) {
        // Install nuget dependencies
        installNugetPackages(for: module, project: projectDir)
        do {
            // Navigate to the directory where the .nuspec file is located
            main.currentdirectory = projectDir.appending("\(module)/\(module)")
            // Run nuget pack command
            try "".runAndPrint("nuget", "pack", "-IncludeReferencedProjects", "-Build", "-Properties", "Configuration=Release")
            //Return to project directory after packing
            main.currentdirectory = projectDir
        } catch {
            Logger.printFatalError("Failed to pack \(module)")
            exit(1)
        }
    }
    
    /**
     Sets API Key for Nuget configuration.
     */
    func setApiKey() {
        let command = main.run(bash: "nuget setapikey \(Constants.xamarinAPIKey)")
        if command.error != nil {
            Logger.printError(command.stderror)
            Logger.printFatalError("Failed to set API Key for publishing Nuget packages!")
        } else {
            Logger.printSuccess("Successfully set API Key for publishing Nuget packages!")
        }
    }
}


class FlutterPublishCommand: Command {
    let name = "publish"
    let shortDescription = "Publishes FLutter Release on Github and Dashboard's changelog. . The following environment variables are required for the command to run: RELEASE_GITHUB_TOKEN, DASHBOARD_TOKEN"
    
    func execute() throws {
        do {
            try publishToGithubAndDashboard()
        } catch {
            print(error)
            print("Publishing to Github and/or dashboard failed");
            exit(1)
        }
    }
    
    private func publishToGithubAndDashboard() throws {
        let owner = "Instabug";
        let repository = "Instabug-Flutter";
        var version = "";
        let changeLog = Helpers.getChangelog();
        let gitToken = ProcessInfo.processInfo.environment["RELEASE_GITHUB_TOKEN"]!;
        let dashboardToken = ProcessInfo.processInfo.environment["DASHBOARD_TOKEN"]!;
        
        // Getting Version number from package.json
        let pathVersion = "pubspec.yaml"
        let dataVersion = try NSString(contentsOfFile: pathVersion,
                                       encoding: String.Encoding.ascii.rawValue)
        
        
        let matched = Helpers.matches(for: "version: ([0-9]-*.*)+[0-9]", in: dataVersion as String);
        let grepVersion = Helpers.matches(for: "([0-9]-*.*)+[0-9]", in: matched! as String);
        version = "v" + grepVersion!;
        
        
        // Releases as bullet points seperated by line breaks (for github release)
        var gitReleaseNotes = "";
        var markdownChangelog = "";
        for change in changeLog {
            gitReleaseNotes += change + "\\n";
            markdownChangelog += change + "\n";
        }
        markdownChangelog += "\n";
        
        // Publish tag and Release to Github
        try "".runAndPrint("curl", "--data", "{\"tag_name\": \"\(version)\", \"target_commitish\": \"master\", \"name\": \"Release \(version)\", \"body\": \"\(gitReleaseNotes)\", \"draft\": false, \"prerelease\": false}", "https://api.github.com/repos/\(owner)/\(repository)/releases?access_token=\(gitToken)");
        
                // Converting markdown to HTML for dashboard changelog
                guard let dashboardChangelogs = markdownChangelog.markdownToHTML else {
                    return ;
                }
                let dashboardChangelog = String(dashboardChangelogs.filter { !"\n".contains($0) });
        
        
                // Getting today's date to be published on the dashboard
                let date = Date();
                let formatter = DateFormatter();
                formatter.dateFormat = "MMM dd yyyy";
                let today = formatter.string(from: date);
        
                // Publishing Release and Changelog to dashboard
                try "".runAndPrint("curl", "-X", "POST", "-H", "content-type: application/json", "-f", "-d", "{\"token\": \"\(dashboardToken)\",\"change_log\": {\"version\": \"\(version)\",\"release_date\": \"\(today)\",\"target_os\": \"8\",\"sdk_changes\": \"\(dashboardChangelog)\"},\"sdk_remote_url\": \"https://github.com/Instabug/Instabug-Flutter\",\"sample_app_remote_url\": \"https://github.com/Instabug/Instabug-Flutter/tree/master/example\"}", "https://api.instabug.com/api/internal/change_logs");
        
    }
}

class CordovaPublishCommand: Command {
    let name = "publish"
    let shortDescription = "Publish the cordova Instabug library to npm, creates a release on Github along with the changelog and updates the dashboard's changelog. The following environment variables are required for the command to run: RELEASE_GITHUB_TOKEN, DASHBOARD_TOKEN, NPM_TOKEN"
    
    func execute() throws {
        do {
            try publishToGithubAndDashboard()
        } catch {
            print(error)
            print("Publishing to Github and/or dashboard failed");
            exit(1)
        }
        do {
            try publishToNpm()
        } catch {
            print(error)
            print("Publishing to npm failed");
            exit(1)
        }
        
    }
    
    private func publishToGithubAndDashboard() throws {
        let owner = "Instabug";
        let repository = "instabug-cordova";
        var version = "";
        let changeLog = Helpers.getChangelog();
        let gitToken = ProcessInfo.processInfo.environment["RELEASE_GITHUB_TOKEN"]!;
        let dashboardToken = ProcessInfo.processInfo.environment["DASHBOARD_TOKEN"]!;
        
        // Getting Version number from package.json
        let pathVersion = "package.json"
        let dataVersion = try NSString(contentsOfFile: pathVersion,
                                       encoding: String.Encoding.ascii.rawValue)
        
        let matched = Helpers.matches(for: "\"version\": \"([0-9]-*.*)+[0-9]\"", in: dataVersion as String)
        let grepVersion = Helpers.matches(for: "([0-9]{1,}\\.)+[0-9]{1,}", in: matched! as String)
        version = "v" + grepVersion!;
        
        // Releases as bullet points seperated by line breaks (for github release)
        var gitReleaseNotes = "";
        var markdownChangelog = "";
        for change in changeLog {
            gitReleaseNotes += change + "\\n";
            markdownChangelog += change + "\n";
        }
        markdownChangelog += "\n";
        
        // Publish tag and Release to Github
        try "".runAndPrint("curl", "--data", "{\"tag_name\": \"\(version)\", \"target_commitish\": \"master\", \"name\": \"Release \(version)\", \"body\": \"\(gitReleaseNotes)\", \"draft\": false, \"prerelease\": false}", "https://api.github.com/repos/\(owner)/\(repository)/releases?access_token=\(gitToken)");
        
        // Converting markdown to HTML for dashboard changelog
        guard let dashboardChangelogs = markdownChangelog.markdownToHTML else {
            return ;
        }
        let dashboardChangelog = String(dashboardChangelogs.filter { !"\n".contains($0) });
        
        
        // Getting today's date to be published on the dashboard
        let date = Date();
        let formatter = DateFormatter();
        formatter.dateFormat = "MMM dd yyyy";
        let today = formatter.string(from: date);
        
        // Publishing Release and Changelog to dashboard
        try "".runAndPrint("curl", "-X", "POST", "-H", "content-type: application/json", "-f", "-d", "{\"token\": \"\(dashboardToken)\",\"change_log\": {\"version\": \"\(version)\",\"release_date\": \"\(today)\",\"target_os\": \"6\",\"sdk_changes\": \"\(dashboardChangelog)\"},\"sdk_remote_url\": \"https://github.com/Instabug/instabug-cordova\",\"sample_app_remote_url\": \"https://github.com/Instabug/instabug-cordova/tree/master/sample\"}", "https://api.instabug.com/api/internal/change_logs");
        
    }
    
    private func publishToNpm() throws {
        // Login to npm
        let token = ProcessInfo.processInfo.environment["NPM_TOKEN"]!;
        let command = main.run(bash: "npm set //registry.npmjs.org/:_authToken=\(token)");
        if command.error != nil {
            Logger.printError(command.stderror);
        } else {
            Logger.printSuccess("Logged in to npm with username instabug");
        }
        
        // Publish to npm
        try "".runAndPrint("npm", "publish", "./");
    }
}
