import Foundation

public class Networking {
    let baseUrl: URL
    let session = URLSession.shared
    
    public init(baseUrl: URL) {
        self.baseUrl = baseUrl
    }
    
    public func getSynchronously(path: String, parameters: [String: Any]) -> NetworkResponse {
        let url = relativeUrlFor(path: path)
        let request = urlEncodedHttpRequest(url: url, method: "GET", parameters: parameters)
        
        return performSynchronousRequest(request: request)
    }
    
    // Ignores base URL
    public func getSynchronously(url: URL, parameters: [String: Any]) -> NetworkResponse {
        let request = urlEncodedHttpRequest(url: url, method: "GET", parameters: parameters)
        return performSynchronousRequest(request: request)
    }
    
    public func postSynchronously(path: String, parameters: [String: Any]) -> NetworkResponse {
        let url = relativeUrlFor(path: path)
        let request = jsonBodyHttpRequest(url: url, method: "POST", parameters: parameters)
        return performSynchronousRequest(request: request)
    }
    
    func performSynchronousRequest(request: URLRequest) -> NetworkResponse {
        let semaphore = DispatchSemaphore(value: 0)
        var networkResponse = NetworkResponse(request: nil, response: nil, data: nil, error: nil)
        
        let task = session.dataTask(with: request) { (data, urlResponse, error) in
            networkResponse.request = request
            networkResponse.response = urlResponse as? HTTPURLResponse
            networkResponse.data = data
            networkResponse.error = error
            
            semaphore.signal()
        }
        
        task.resume()
        semaphore.wait()
        
        return networkResponse
    }
    
    func relativeUrlFor(path: String) -> URL {
        let url = baseUrl.appendingPathComponent(path)
        return url
    }
    
    func urlEncodedHttpRequest(url: URL, method: String, parameters: [String: Any]) -> URLRequest {
        let query = queryStringFor(parameters: parameters)
        let encodedURL = url.appending(query: query)
        
        var urlRequest = URLRequest(url: encodedURL)
        urlRequest.httpMethod = method
        
        return urlRequest
    }
    
    func jsonBodyHttpRequest(url: URL, method: String, parameters: [String: Any]) -> URLRequest {
        var urlRequest = URLRequest(url: url)
        let jsonData = try? JSONSerialization.data(withJSONObject: parameters)
        urlRequest.httpBody = jsonData
        urlRequest.httpMethod = method
        urlRequest.setValue("application/json; charset=utf-8", forHTTPHeaderField: "Content-Type")
        return urlRequest
    }
    
    func queryStringFor(parameters: [String: Any]) -> String {
        var urlParameters = [String]()

        for (key, value) in parameters {
            guard let stringValue = value as? String else {
                continue
            }

            var characters = NSCharacterSet.urlQueryAllowed
            characters.remove(charactersIn: "&")

            if var encodedValue = stringValue.addingPercentEncoding(withAllowedCharacters: characters) {
                // Some requests need to percent-encode the + sign, while others don't. So I had to
                // create a special literal string for it ¯\_(ツ)_/¯
                encodedValue = encodedValue.replacingOccurrences(of: "!plusliteral", with: "%2B")

                let parameters = "\(key)=\(encodedValue)"
                urlParameters.append(parameters)
            }
        }

        return urlParameters.joined(separator: "&")
    }

}
